import os
from datetime import datetime
import numpy as np


def is_recent(fname):
    lm = os.path.getmtime(fname)
    lm = datetime.utcfromtimestamp(lm)
    now = datetime.utcnow()
    dt = (now - lm).total_seconds()
    recent = True if dt < 86400 else False
    return recent


def _time(t):
    if isinstance(t, datetime):
        return t
    elif isinstance(t, str) and bool(re.match("\d{4}-\d\d-\d\d \d\d:\d\d:\d\d", t)):
        return datetime.strptime(t, '%Y-%m-%d %H:%M:%S')
    elif t is None:
        return datetime.utcnow()
    else:
        raise Exception("only datetime or str: '%Y-%m-%d %H:%M:%S'")


def mag(x):
    """Returns magnitude of a vector"""
    return np.linalg.norm(np.array(x))

def unit(x):
    """Returns unit vector of a vector"""
    return x / mag(x)

def rev(x):
    return x % 360


def _in_vec(inp):
    if len(inp.shape)>1:
        if inp.shape[-1]==3:
            vec_out = inp[:,0], inp[:,1], inp[:,2]
        elif inp.shape[-1]==2:
            vec_out = inp[:,0], inp[:,1]
        else:
            raise Exception('Input imension not sopported!')
    elif len(inp.shape)==1:
        vec_out = inp
    else:
        raise Exception('Dimensionality problem! 2')
    return vec_out


def _out_vec(vec):
    if len(vec.shape)>1:
        vec = vec.T
        if vec.shape[-1]==3:
            x, y, z = vec[:,0], vec[:,1], vec[:,2]
            vec_out = np.vstack((x, y, z)).T
        elif vec.shape[-1]==2:
            x, y = vec[:,0], vec[:,1]
            vec_out = np.vstack((x, y)).T
        else:
            raise Exception('Dimensionality problem! 3')
    elif len(vec.shape)==1:
        vec_out = vec
    else:
        raise Exception('Dimensionality problem! 4')
    return vec_out



"""
Module iau
===============
Transform coordinates from GCRS to True Equator True Equinox and many more
IAU 2000A models
Ref: https://arxiv.org/abs/astro-ph/0602086
"""

import csv
import os.path
from urllib.request import urlretrieve
import numpy as np
from datetime import datetime, timedelta
from .time import utc2tdb, datetime_to_jd
from .utils import is_recent

NUT_URL = 'https://raw.githubusercontent.com/behrouzz/astrodatascience/main/data/iau2000A_nut.csv'
FINALS2000A_URL = 'https://maia.usno.navy.mil/ser7/finals2000A.all'
FINALS2000A_FNAME = 'finals2000A.all'

s2r = (1/3600) * (np.pi/180)
d2r = np.pi / 180

def get_T(utc):
    t = utc2tdb(utc)
    T = (datetime_to_jd(t) - 2451545) / 36525
    return T

def precession(T):
    e0 = 84381.406
    psi_A_cf = [0, 5038.481507, -1.0790069, -0.00114045, 0.000132851, -0.0000000951]
    w_A_cf = [e0, -0.025754, 0.0512623, -0.00772503, -0.000000467, 0.0000003337]
    chi_A_cf = [0, 10.556403, -2.3814292, -0.00121197, 0.000170663, -0.0000000560]

    psi_A = np.poly1d(psi_A_cf[::-1])(T)
    w_A = np.poly1d(w_A_cf[::-1])(T)
    chi_A = np.poly1d(chi_A_cf[::-1])(T)

    # convert to radian
    e0 = e0 * s2r
    psi_A = psi_A * s2r
    w_A = w_A * s2r
    chi_A = chi_A * s2r

    s1 = np.sin(e0)
    s2 = np.sin(-psi_A)
    s3 = np.sin(-w_A)
    s4 = np.sin(chi_A)

    c1 = np.cos(e0)
    c2 = np.cos(-psi_A)
    c3 = np.cos(-w_A)
    c4 = np.cos(chi_A)

    row1 = [c4*c2-s2*s4*c3, c4*s2*c1+s4*c3*c2*c1-s1*s4*s3, c4*s2*s1+s4*c3*c2*s1+c1*s4*s3]
    row2 = [-s4*c2-s2*c4*c3, -s4*s2*c1+c4*c3*c2*c1-s1*c4*s3, -s4*s2*s1+c4*c3*c2*s1+c1*c4*s3]
    row3 = [s2*s3, -s3*c2*c1-s1*c3, -s3*c2*s1+c3*c1]

    P = np.array([row1, row2, row3])
    return P




def create_phi(T): #new
    fname = 'iau2000A_nut.csv'
    if not os.path.exists(fname):
        print('Downloading '+fname+'...')
        urlretrieve(NUT_URL, fname)

    data = []
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter=',')
        for row in reader:
            data.append(row)
    data = data[1:]
    
    M = np.array([[int(i) for i in row[1:15]] for row in data])
    S = np.array([float(i[15]) for i in data]) * s2r
    Sd = np.array([float(i[16]) for i in data]) * s2r
    Cp = np.array([float(i[17]) for i in data]) * s2r
    C = np.array([float(i[18]) for i in data]) * s2r
    Cd = np.array([float(i[19]) for i in data]) * s2r
    Sp = np.array([float(i[20]) for i in data]) * s2r

    phi = {
    1 : 908103.259872 + 538101628.688982 * T,
    2 : 655127.283060 + 210664136.433548 * T,
    3 : 361679.244588 + 129597742.283429 * T,
    4 : 1279558.798488 + 68905077.493988 * T,
    5 : 123665.467464 + 10925660.377991 * T,
    6 : 180278.799480 + 4399609.855732 * T,
    7 : 1130598.018396 + 1542481.193933 * T,
    8 : 1095655.195728 + 786550.320744 * T,
    9 : 5028.8200 * T + 1.112022 * T**2,
    10 : 485868.249036 + 1717915923.2178 * T + 31.8792 * T**2 + 0.051635 * T**3 - 0.00024470 * T**4,
    11 : 1287104.79305 + 129596581.0481 * T - 0.5532 * T**2 + 0.000136 * T**3 - 0.00001149 * T**4,
    12 : 335779.526232 + 1739527262.8478 * T - 12.7512 * T**2 - 0.001037 * T**3 + 0.00000417 * T**4,
    13 : 1072260.70369 + 1602961601.2090 * T - 6.3706 * T**2 + 0.006593 * T**3 - 0.00003169 * T**4,
    14 : 450160.398036 - 6962890.5431 * T + 7.4722 * T**2 + 0.007702 * T**3 - 0.00005939 * T**4,
    }


    d_psi = 0
    d_e = 0
    for i in range(1,1366):
        PHI = 0
        for j in range(1,15):
            PHI = PHI + M[i-1,j-1] * phi[j]
        PHI = PHI * s2r
        d_psi = d_psi + ((S[i-1]+Sd[i-1]*T)*np.sin(PHI) + Cp[i-1]*np.cos(PHI))
        d_e = d_e + ((C[i-1]+Cd[i-1]*T)*np.cos(PHI) + Sp[i-1]*np.sin(PHI))

    e0 = 84381.406
    e = e0 - 46.836769*T - 0.0001831 *T**2 + 0.00200340 *T**3 - 0.000000576 *T**4 - 0.0000000434 *T**5
    e = e * s2r

    F = phi[12] * s2r
    D = phi[13] * s2r
    om = phi[14] * s2r
    return d_psi, d_e, e, F, D, om

    

def nutation(d_psi, d_e, e):
    #d_psi, d_e, e, _, _, _ = create_phi(T)
    S1 = np.sin(e)
    S2 = np.sin(-d_psi)
    S3 = np.sin(-e-d_e)
    C1 = np.cos(e)
    C2 = np.cos(-d_psi)
    C3 = np.cos(-e-d_e)
    row1 = [C2, S2*C1, S2*S1]
    row2 = [-S2*C3, C3*C2*C1-S1*S3, C3*C2*S1+C1*S3]
    row3 = [S2*S3, -S3*C2*C1-S1*C3, -S3*C2*S1+C3*C1]
    return np.array([row1, row2, row3])


def equation_of_equinoxes(T, d_psi, e, F, D, om):
    # Equation of equinoxes (arcsec)
    eq_eq = d_psi * np.cos(e) \
        + 0.00264096 * np.sin(om) \
        + 0.00006352 * np.sin(2*om) \
        + 0.00001175 * np.sin(2*F - 2*D + 3*om) \
        + 0.00001121 * np.sin(2*F - 2*D + om) \
        - 0.00000455 * np.sin(2*F - 2*D + 2*om) \
        + 0.00000202 * np.sin(2*F + 3*om) \
        + 0.00000198 * np.sin(2*F + om) \
        - 0.00000172 * np.sin(3*om) \
        - 0.00000087 * T * np.sin(om)
    return eq_eq


def equation_of_origins(T, eq_eq):
    par = 0.014506 + 4612.156534*T + 1.3915817* T**2 \
          - 0.00000044 *T**3 - 0.000029956 *T**4 - 0.0000000368 *T**5
    return -(par + eq_eq)

# B : frame bias matrix
mas2rad = 4.84813681109536e-09

da0 = -14.6 * mas2rad
e0 = -16.6170 * mas2rad
n0 = -6.8192 * mas2rad

B = np.array([[ 1-0.5*(da0**2 + e0**2), da0, -e0],
              [-da0-n0*e0, 1-0.5*(da0**2 + n0**2), -n0],
              [e0-n0*da0, n0+e0*da0, 1-0.5*(n0**2+e0**2)]])

def tete_rotmat(t):
    T = get_T(t)
    P = precession(T)
    d_psi, d_e, e, _, _, _ = create_phi(T)
    N = nutation(d_psi, d_e, e)
    rot_mat = np.matmul(np.matmul(N, P), B)
    return rot_mat

def gcrs2tete(pos, t):
    rot_mat = tete_rotmat(t)
    return np.matmul(rot_mat, pos)


def get_C(T):
    """
    Get C matrix

    T can be calculated with this formula:
    T = (datetime_to_jd(utc2tdb(t)) - 2451545) / 36525

    Using C matrix, one can tranfsorm from GCRS to CIRS:
    r_CIRS = C * r_GCRS

    Arguments:
    ----------
        T: number of centuries of TDB from J2000

    Return:
    -------
        C matrix    
    """
    P = precession(T)
    d_psi, d_e, e, F, D, om = create_phi(T)
    N = nutation(d_psi, d_e, e)
    
    tmp = np.matmul(np.matmul(B.T, P.T), N.T)
    n_GCRS = np.matmul(tmp, np.array([0,0,1]))
    eq_GCRS = np.matmul(tmp, np.array([1,0,0]))

    eq_eq = equation_of_equinoxes(T, d_psi, e, F, D, om)
    eq_or = equation_of_origins(T, eq_eq)
    eq_o_rad = eq_or * s2r

    # sig_GCRS: position of CIO in GCRS
    sig_GCRS = eq_GCRS * np.cos(eq_o_rad) - np.cross(n_GCRS, eq_GCRS) * np.sin(eq_o_rad)
    y_GCRS = np.cross(n_GCRS, sig_GCRS) # page 65

    C = np.array([sig_GCRS, y_GCRS, n_GCRS])
    # page 66 and 48 (MOHEM: combined method)
    return C


def loc2itrs(lon, lat, h):
    LON, LAT = lon*d2r, lat*d2r
    a = 6378137
    f = 1 / 298.257223563
    C = 1 / np.sqrt(np.cos(LAT)**2 + (1-f)**2 * np.sin(LAT)**2)
    S = (1 - f)**2 * C
    r = np.array([(a*C+h)*np.cos(LAT)*np.cos(LON),
                  (a*C+h)*np.cos(LAT)*np.sin(LON),
                  (a*S+h)*np.sin(LAT)])
    return r


def interpolate(t, array):
    mjd = datetime_to_jd(t) - 2400000.5
    m1 = mjd>=array[:,0]
    m2 = mjd<array[:,0]
    t1 = array[m1][-1] # previous
    t2 = array[m2][0] # next
    # interpolate
    cf = np.polyfit([t1[0],t2[0]], [t1[1],t2[1]], 1)
    f = np.poly1d(cf)
    return f(mjd)


def get_finals2000a():
    should_download = True
    if os.path.exists(FINALS2000A_FNAME):
        if is_recent(FINALS2000A_FNAME):
            should_download = False
    if should_download:
        print('Downloading '+FINALS2000A_FNAME+'...')
        urlretrieve(FINALS2000A_URL, FINALS2000A_FNAME)

    with open(FINALS2000A_FNAME, 'r') as f:
        data = f.read().split('\n')

    data = [i for i in data if len(i.strip())>68]
    mjd = [int(i[7:12]) for i in data]
    dut1 = [float(i[58:68]) for i in data]
    pm_x = [float(i[18:27]) for i in data]
    pm_y = [float(i[37:46]) for i in data]

    dut1_array = np.array(list(zip(mjd,dut1)))
    pm_x = np.array(list(zip(mjd, pm_x)))
    pm_y = np.array(list(zip(mjd, pm_y)))
    return dut1_array, pm_x, pm_y


def ut1_utc(t, dut1_array):
    mjd = datetime_to_jd(t) - 2400000.5
    m1 = mjd>=dut1_array[:,0]
    m2 = mjd<dut1_array[:,0]
    t1 = dut1_array[m1][-1] # previous
    t2 = dut1_array[m2][0] # next
    # interpolate
    cf = np.polyfit([t1[0],t2[0]], [t1[1],t2[1]], 1)
    f = np.poly1d(cf)
    return f(mjd)


def get_W(t, pm_x, pm_y):
    x = interpolate(t, pm_x) * s2r
    y = interpolate(t, pm_y) * s2r
    #T = get_T(t)
    #s_p = -47 * microascsec * T

    row1 = [np.cos(x), np.sin(x)*np.sin(y), -np.sin(x)*np.cos(y)]
    row2 = [0, np.cos(y), np.sin(y)]
    row3 = [np.sin(x), -np.sin(y), np.cos(x)*np.cos(y)]
    W = np.array([row1, row2, row3])
    # or: W = np.array([[1, 0, -x], [0, 1, y], [x, -y, 1]])
    return W

def R1(x):
    r1 = [1, 0, 0]
    r2 = [0, np.cos(x), np.sin(x)]
    r3 = [0, -np.sin(x), np.cos(x)]
    return np.array([r1, r2, r3])

def R2(x):
    r1 = [np.cos(x), 0, -np.sin(x)]
    r2 = [0, 1, 0]
    r3 = [np.sin(x), 0, np.cos(x)]
    return np.array([r1, r2, r3])

def R3(x):
    r1 = [np.cos(x), np.sin(x), 0]
    r2 = [-np.sin(x), np.cos(x), 0]
    r3 = [0, 0, 1]
    return np.array([r1, r2, r3])

def get_ERA(utc, dut1):
    #dut1_arr, _, _ =get_finals2000a()
    #dut1 = ut1_utc(utc, dut1_arr)
    ut1 = utc + timedelta(seconds=dut1)
    Du = datetime_to_jd(ut1) - 2451545
    ERA = 2*np.pi * (0.7790572732640 + 1.00273781191135448 * Du)
    era_deg = ERA * (180/np.pi)
    return era_deg % 360


def get_era_eqor(ra, dec, obs_loc, t):
    """
    Calculate hour angle of an object

    Arguments
    ---------
        obs_loc (tuple): (longtitude, latitude) of observer
        ra (float): RA of the object (equinox of date)
        dec (float): DEC of the object (equinox of date)
        t (datetime): time of observation in UTC        

    Returns
    -------
        hour angle (degrees)
    """
    LON, LAT = obs_loc
    T = get_T(t)
    
    # Calculate ERA
    # =============
    dut1_array, pm_x, pm_y = get_finals2000a()
    dut1 = ut1_utc(t, dut1_array)
    ERA = get_ERA(t, dut1)

    # Find equation of origins
    d_psi, d_e, e, F, D, om = create_phi(T)
    eq_eq = equation_of_equinoxes(T, d_psi, e, F, D, om)
    eq_or = equation_of_origins(T, eq_eq)

    # Real lon and lat
    # ================
    #lon , lat = LON, LAT
    xp = interpolate(t, pm_x)
    yp = interpolate(t, pm_y)
    lon = LON + (xp*np.sin(LON*d2r) + yp*np.cos(LON*d2r)) * np.tan(LAT*d2r) / 3600
    lat = LAT + (xp*np.cos(LON*d2r) - yp*np.sin(LON*d2r)) / 3600

    # Calculate hour angle of object
    # ==============================
    # ra     : RA of object wrt equinox of date
    # ra_sig : RA of object wrt CIO
    eq_or = eq_or / 3600 # convert to degrees
    #ra_sig = (ra + eq_or) % 360
    #ha = (ERA - ra_sig + lon) % 360
    return ERA, eq_or, (lon, lat)


def get_ha(ra, dec, obs_loc, t):
    """
    Calculate hour angle of an object

    Arguments
    ---------
        obs_loc (tuple): (longtitude, latitude) of observer
        ra (float): RA of the object (equinox of date)
        dec (float): DEC of the object (equinox of date)
        t (datetime): time of observation in UTC        

    Returns
    -------
        hz (deg), obs_loc (lon, lat)
    """
    
    era, eqor, obs_loc = get_era_eqor(ra, dec, obs_loc, t)
    lon = obs_loc[0]
    ra_sig = (ra + eqor) % 360
    ha = (era - ra_sig + lon) % 360
    return ha, obs_loc


"""
T = get_T(t)
GMST = 86400 * ERA + (0.014506 + 4612.156534*T + 1.3915817* T**2 \
         - 0.00000044 *T**3 - 0.000029956 *T**4 - 0.0000000368 *T**5)/15
GAST = GMST + equation_of_equinoxes/15
LAST = GAST + (3600/15)*lon # lon should be calculated with eq page 17
"""



"""
Module horizons
===============
This module has two classes (Vector and Observer) to return positions
of the solar sytem objects by querying NASA's JPL Horizons API.

The Vector class provides positions with three dimensions: x,y,z.

The Observer class provides positions with two dimensions: ra,dec or
alt,az; depending on the request.
"""

import re
from datetime import datetime, timedelta
from urllib.request import urlopen
from requests import request
import numpy as np
import pandas as pd
from .plots import plot_altaz, plot_xyz
from .data.cities import cities
import matplotlib.pyplot as plt


BASE_URL = 'https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&'


def _time_format(t): #check time format
    correct = True
    if isinstance(t, datetime):
        t = t.isoformat()[:19]
    elif isinstance(t, str) and bool(re.match("\d{4}-\d\d-\d\d \d\d:\d\d:\d\d", t)):
        t = t.replace(' ', 'T')
    else:
        correct = False
    return [correct, t]


def download(script):
    url = BASE_URL + script
    error_msg = ''
    req = request('GET', url)
    all_text = req.content.decode('utf-8')
    if ('$$SOE' not in all_text) or ('$$EOE' not in all_text):
        error_msg = all_text[:all_text.find('$$SOF')]
        #return error_msg
        print(error_msg)
    mark1 = all_text.find('$$SOE')
    text = all_text[mark1+6:]
    mark2 = text.find('$$EOE')
    text = text[:mark2]
    raw_rows = text.split('\n')[:-1]
    n_columns = len(raw_rows[0].split(',')[:-1])

    head = all_text[:mark1]
    cols = re.findall(r'[a-zA-Z()_\s]+,', head)
    cols = "".join(cols)
    cols = cols.split('\n')[-1].split(',')
    cols = [i.strip() for i in cols][:-1]

    raw_rows = [i.split(',')[:-1] for i in raw_rows]
    rows = []
    for r in raw_rows:
        row = [i.strip() for i in r]
        rows.append(row)
    df = pd.DataFrame(data=rows, columns=cols)
    return modify_jpl_df(df)


def modify_jpl_df(df):
    df['Calendar Date (TDB)'] = \
                 df['Calendar Date (TDB)'].str[5:]\
                 .apply(lambda i: datetime.strptime(i, '%Y-%b-%d %H:%M:%S.%f'))
    df = df.set_index('Calendar Date (TDB)')
    df.index.name = 'TBD'
    df = df.astype(float)
    return df


class Vector:
    """
    Vector type coordinates (x, y, z)
    
    Parameters
    ----------
        target (str): targer body
        t1 (datetime or str): time from which to request data
        t2 (datetime or str): time to which to request data
        step (int): number of time intervals
        center (str): origin of coordinates. format: site@body
                      default '500@0'
        ref_plane (str): refrence plane: 'FRAME', 'ECLIPTIC' or 'BODY EQUATOR'
                         default 'FRAME' (equator plane)
        vec_table (int): vector table type: 1 for position; 2 for position and velocity
    
    Attributes
    ----------
        time (list): list of datetime objects; each element is the time
                     at which the coordinates is presented
        pos (np.array): coordinates array with three columns (x, y, z)
                        shape of array is (n, 3); n: number of steps
        x (np.array): x component of position vector
        y (np.array): y component of position vector
        z (np.array): z component of position vector
        vel (np.array): velocity array with three columns (vx, vy, vz)
                        shape of array is (n, 3); n: number of steps
        vx (np.array): x component of velocity vector
        vy (np.array): y component of velocity vector
        vz (np.array): z component of velocity vector
        url (str): url of NASA's JPL Horizons that produced the results
    
    Methods
    -------
        plot: plot position(s) (x,y,z) of the target body
    """

    def __init__(self, target, t1, t2=None, step=None, center='500@0', ref_plane='FRAME', vec_table=1):
        self.target = target
        self.center = center
        self.ref_plane = ref_plane
        self.vec_table = vec_table
        
        correct, t1 = _time_format(t1)
        if correct:
            self.t1 = t1
        else:
            raise Exception("t1 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
        self.init = False
        if t2 is None:
            self.init = True # user just wants the initial state
            step = 1
            t2 = datetime.strptime(self.t1, '%Y-%m-%dT%H:%M:%S') + timedelta(seconds=1)
        
        correct, t2 = _time_format(t2)
        if correct:
            self.t2 = t2
        else:
            raise Exception("t2 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
            
        self.step = step

        error_msg, time, pos, vel = self.get_request()
        
        if len(error_msg)==0:
            if self.init:
                self.time = time[0]
                self.pos = pos[0]
                self.x, self.y, self.z = pos[0][0], pos[0][1], pos[0][2]
                if vel.shape!=(0,):
                    self.vel = vel[0]
                    self.vx, self.vy, self.vz = vel[0,0], vel[0,1], vel[0,2]
                else:
                    self.vel = vel
                
            else:
                self.time = time
                self.pos = pos
                self.x, self.y, self.z = pos[:,0], pos[:,1], pos[:,2]
                self.vel = vel
                if vel.shape!=(0,):
                    self.vx, self.vy, self.vz = vel[:,0], vel[:,1], vel[:,2]
                else:
                    self.vx, self.vy, self.vz = vel, vel, vel
        else:
            raise Exception('\n'+error_msg[:-2])

    def vector_url(self, target, t1, t2, step, center, ref_plane, vec_table):
        params = f"""COMMAND='{target}'
        CENTER='{center}'
        MAKE_EPHEM='YES'
        TABLE_TYPE='VECTORS'
        START_TIME='{t1}'
        STOP_TIME='{t2}'
        STEP_SIZE='{step}'
        CAL_FORMAT='CAL'
        TIME_DIGITS='FRACSEC'
        OUT_UNITS='KM-S'
        REF_PLANE='{ref_plane}'
        REF_SYSTEM='J2000'
        VECT_CORR='NONE'
        VEC_LABELS='NO'
        VEC_DELTA_T='NO'
        CSV_FORMAT='YES'
        OBJ_DATA='NO'
        VEC_TABLE='{vec_table}'"""
        params = params.replace('\n', '&').replace(' ', '')
        url = BASE_URL + params
        self.url = url
        return url

    def get_request(self):
        error_msg = ''
        url = self.vector_url(self.target, self.t1, self.t2, self.step,
                              self.center, self.ref_plane, self.vec_table)
        with urlopen(url) as r:
            text = r.read().decode('utf-8')
        if ('$$SOE' not in text) or ('$$EOE' not in text):
            error_msg = text[:text.find('$$SOF')]
            return [error_msg, np.array([]), np.array([])]
        mark1 = text.find('$$SOE')
        text = text[mark1+6:]
        mark2 = text.find('$$EOE')
        text = text[:mark2]
        rows = text.split('\n')[:-1]
        n_columns = len(rows[0].split(',')[:-1])
        
        times = [i.split(',')[1].strip()[5:] for i in rows]
        times = [datetime.strptime(i, '%Y-%b-%d %H:%M:%S.%f') for i in times]

        # positions
        pS = [i.split(',')[2:5] for i in rows]
        pos_list = []
        for p in pS:
            tmp_ls = [float(i.strip())*1000 for i in p]
            pos_list.append(tmp_ls)

        # velocities
        if n_columns==8:
            vS = [i.split(',')[5:8] for i in rows]
            vel_list = []
            for v in vS:
                tmp_ls = [float(i.strip())*1000 for i in v]
                vel_list.append(tmp_ls)
        else:
            vel_list = []
        return [error_msg, times, np.array(pos_list), np.array(vel_list)]

    def plot(self, color='b', size=10):
        """
        plot position(s) (ra/dec or az/alt) of the target body
        
        Arguments
        ----------
            color (str): color of body
            size (int): size of body
        
        Returns
        -------
            matplotlib.axes.Axes object
        """
        if isinstance(self.time, list):
            return plot_xyz(self.x, self.y, self.z, color, size)
        else:
            return plot_xyz([self.x], [self.y], [self.z], color, size)


class Observer:
    """
    Observer type coordinates (ra/dec or az/alt)
    
    Parameters
    ----------
        ttarget (str): targer body
        t1 (datetime or str): time from which to request data
        t2 (datetime or str): time to which to request data
        step (int): number of time intervals
        center (str): origin of coordinates. format: site@body
                      default '500@0'
        quantities (int): 1 for ra/dec, 2 for az/alt; (default 2)
    
    Attributes
    ----------
        time (list): list of datetime objects; each element is the time
                     at which the coordinates is presented
        pos (np.array): coordinates array with two columns (ra/dec or az/alt)
                        shape of array is (n, 2); n: number of steps
        ra (np.array): right ascension OR azimuth (depending on quantities chosen)
        dec (np.array): declination OR altitude (depending on quantities chosen)
        url (str): url of NASA's JPL Horizons that produced the results
    
    Methods
    -------
        plot: plot position(s) (ra/dec or az/alt) of the target body
    """
    
    def __init__(self, target, t1, t2=None, step=None, center='500@399', quantities=2):
        self.target = target
        self.center = center
        self.quantities = quantities
        
        correct, t1 = _time_format(t1)
        if correct:
            self.t1 = t1
        else:
            raise Exception("t1 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
        self.init = False
        if t2 is None:
            self.init = True # user just wants the initial state
            step = 1
            t2 = datetime.strptime(self.t1, '%Y-%m-%dT%H:%M:%S') + timedelta(seconds=1)
        
        correct, t2 = _time_format(t2)
        if correct:
            self.t2 = t2
        else:
            raise Exception("t2 must be datetime or str: '%Y-%m-%d %H:%M:%S'")
            
        self.step = step

        error_msg, time, pos = self.get_request()
        
        if len(error_msg)==0:
            if self.init:
                self.time = time[0]
                self.pos = pos[0]
                self.ra, self.dec = pos[0][0], pos[0][1]
            else:
                self.time = time
                self.pos = pos
                self.ra, self.dec = pos[:,0], pos[:,1]
        else:
            raise Exception('\n'+error_msg[:-2])

    def coord(self, center):        
        valid_center = True
        str_center = ''

        if len(center.split('@'))==2:
            body = center.split('@')[-1]
        elif len(center.split('@'))==1:
            body = '399'
        else:
            valid_center = False

        crd = center.split('@')[0]
        if len(crd.split(','))==3:
            crd = ','.join(crd.split(',')[:-1]) + ',' + str(int(crd.split(',')[-1])/1000)
        elif len(crd.split(','))==2:
            crd = crd + ',0'
        else:
            valid_center = False

        if valid_center:
            str_center = f"coord@{body}'&COORD_TYPE='GEODETIC'&SITE_COORD='{crd}"

        return [valid_center, str_center]
    
    def observer_url(self, target, t1, t2, step, center, quantities, skip_daylight=True):
        if len(center.split(',')) > 1: #coordinates
            valid_center, center = self.coord(center)
            skip_daylight = False
            quantities = 4 # apparent Ra & DEC
            if not valid_center:
                raise Exception('Center is not valid!')

        skip_daylight='YES' if skip_daylight==True else 'NO'
        
        params = f"""COMMAND='{target}'
        CENTER='{center}'
        MAKE_EPHEM='YES'
        TABLE_TYPE='OBSERVER'
        START_TIME='{t1}'
        STOP_TIME='{t2}'
        STEP_SIZE='{step}'
        CAL_FORMAT='CAL'
        TIME_DIGITS='FRACSEC'
        ANG_FORMAT='DEG'
        OUT_UNITS='KM-S'
        RANGE_UNITS='KM'
        APPARENT='AIRLESS'
        SUPPRESS_RANGE_RATE='YES'
        SKIP_DAYLT='{skip_daylight}'
        EXTRA_PREC='YES'
        R_T_S_ONLY='NO'
        REF_SYSTEM='J2000'
        CSV_FORMAT='YES'
        OBJ_DATA='NO'
        QUANTITIES='{quantities}'"""
        
        params = params.replace('\n', '&').replace(' ', '')
        url = BASE_URL + params
        self.url = url
        
        return url

    def get_request(self):
        error_msg = ''
        url = self.observer_url(self.target, self.t1, self.t2, self.step,
                                self.center, self.quantities)
        with urlopen(url) as r:
            text = r.read().decode('utf-8')
        if ('$$SOE' not in text) or ('$$EOE' not in text):
            error_msg = text[:text.find('$$SOF')]
            return [error_msg, np.array([]), np.array([])]
        mark1 = text.find('$$SOE')
        text = text[mark1+6:]
        mark2 = text.find('$$EOE')
        text = text[:mark2]
        rows = text.split('\n')[:-1]
        times = [i.split(',')[0].strip() for i in rows]
        times = [datetime.strptime(i, '%Y-%b-%d %H:%M:%S.%f') for i in times]
        pS = [i.split(',')[-3:-1] for i in rows]
        pos_list = []
        for p in pS:
            tmp_ls = [float(i.strip()) for i in p]
            pos_list.append(tmp_ls)
        return [error_msg, times, np.array(pos_list)]

    def plot(self, color='b', size=10):
        """
        plot position(s) (ra/dec or az/alt) of the target body
        
        Arguments
        ----------
            color (str): color of body
            size (int): size of body
        
        Returns
        -------
            matplotlib.axes.Axes object
        """
        plot_altaz(self.ra, self.dec, color, size)


def jupiter_moons(loc, t):
    """
    Draw Jupiter and it's Galilean moons
    
    Arguments
    ---------
        loc (str): location of observer: "lon,lat,elevation"
        t (datetime or str): time of observation (UTC)

    Returns
    -------
        fig, ax
    """
    
    if (type(loc)==str) and loc in cities.keys():
        loc = str(cities[loc]).replace(' ','')[1:-1] + '@399'
    else:
        loc = loc.replace(' ','')  + '@399'

    jup = Observer(599, t, center=loc)
    io  = Observer(501, t, center=loc)
    eu  = Observer(502, t, center=loc)
    ga  = Observer(503, t, center=loc)
    ca  = Observer(504, t, center=loc)

    t = jup.time

    fig, ax = plt.subplots()

    ax.scatter(jup.pos[0], jup.pos[1], label='JUP', c='orange', s=200, alpha=0.5)
    ax.scatter(io.pos[0], io.pos[1], label='IO', c='k', s=20)
    ax.scatter(eu.pos[0], eu.pos[1], label='EU', c='b', s=20)
    ax.scatter(ga.pos[0], ga.pos[1], label='GA', c='r', s=20)
    ax.scatter(ca.pos[0], ca.pos[1], label='CA', c='g', s=20)

    text = "Jupiter's Galilean moons | " + t.isoformat().replace('T', ' ')[:16]
    ax.set_title(text)
    ax.ticklabel_format(useOffset=False)
    ax.set_xlabel('Az')
    ax.set_ylabel('Alt')
    plt.legend()
    plt.grid()
    return fig, ax



class RAhms:
    def __init__(self, h=0, m=0, s=0):
        self.h = h
        self.m = m
        self.s = s
        self.hours = h + m/60 + s/3600
        self.minutes = h*60 + m + s/60
        self.seconds = h*3600 + m*60 + s
        self.deg = 15 * self.hours
        self.rad = self.deg * (3.141592653589793/180)
        

class DECdms:
    def __init__(self, sign='+', d=0, m=0, s=0):
        self.sign = sign
        sgn = -1 if sign=='-' else 1
        self.d = d
        self.m = m
        self.s = s
        self.degrees = (d + m/60 + s/3600) * sgn
        self.minutes = (d*60 + m + s/60) * sgn
        self.seconds = (d*3600 + m*60 + s) * sgn
        self.deg = self.degrees
        self.rad = self.deg * (3.141592653589793/180)


class RAdeg:
    def __init__(self, ra):
        self.rad = ra * (3.141592653589793/180)
        self.h = int(ra/15)
        self.m = int(((ra/15)-self.h)*60)
        self.s = ((((ra/15)-self.h)*60)-self.m)*60
        self.str = str(self.h)+'h' + str(self.m)+'m' + str(self.s)+'s'
        

class DECdeg:
    def __init__(self, dec):
        self.rad = dec * (3.141592653589793/180)
        self.sign = '-' if dec<0 else '+'
        dec = abs(dec)
        self.d = int(dec)
        self.m = abs(int((dec-self.d)*60))
        self.s = (abs((dec-self.d)*60)-self.m)*60
        self.str = self.sign + str(self.d)+'d' + \
                   str(self.m)+'m' + str(self.s)+'s'



import pickle, io
import numpy as np
from .time import utc2tt, datetime_to_jd
from datetime import datetime

d2r = np.pi/180
r2d = 180/np.pi

def rev(x):
    if x >= 360:
        while x >= 360:
            x = x - 360
    elif x < 0:
        while x < 0:
            x = 360 + x
    return x


def get_sun(t):
    """
    Get geocentric position of the Sun

    Approximation based on method presented by The Astronomical Almanac,
    to calculate the apparent coordinates of the Sun, mean equinox and
    ecliptic of date, to a precision of about 0°.01 (36″), for dates 
    between 1950 and 2050.
    Ref: https://en.wikipedia.org/wiki/Position_of_the_Sun

    Parameters
    ----------
        t (datetime): time of observation (UTC)

    Returns
    -------
        ra  : RA of the Sun (deg)
        dec : DEC of the Sun (deg)
        r   : Distance of the Sun (AU)
    """
    tt = utc2tt(t)
    jd = datetime_to_jd(tt)
    n = jd - 2451545
    L = rev(280.460 + 0.985647*n) # mean longitude
    g = rev(357.528 + 0.9856003*n) # mean anomaly
    # Ecliptic coordinates
    ecl_lon = rev(L + 1.915 * np.sin(g*d2r) + 0.020 * np.sin(2*g*d2r))
    ecl_lat = 0
    r = 1.00014 - 0.01671*np.cos(g*d2r) - 0.00014*np.cos(2*g*d2r)
    # Equatorial coordinates
    obl = 23.439 - 0.0000004*n
    ra = rev(np.arctan2(np.cos(obl*d2r)*np.sin(ecl_lon*d2r), np.cos(ecl_lon*d2r))*r2d)
    dec = np.arcsin(np.sin(obl*d2r)*np.sin(ecl_lon*d2r))*r2d
    return ra, dec, r


def dec_sun(t):
    t = utc2tt(t)
    t0 = datetime(t.year, 1, 1)
    dt = t - t0
    N = dt.days + dt.seconds/86400 + dt.microseconds/(1_000_000*86400)
    in_cos = 0.98565*d2r*(N+10) + 1.914*d2r * np.sin(0.98565*d2r*(N-2))
    dec = -np.arcsin(0.39779 * np.cos(in_cos)) * r2d
    return dec



def jd_to_sec(jd):
    """jd to seconds since J2000"""
    return (jd - 2451545.0) * 86400.0


def num2txt(arr):
    output = io.BytesIO()
    np.savetxt(output, arr)
    return output.getvalue().decode('utf-8')


class Segment:
    """
    Segment of SPK data created by numeph python package
    """
    def __init__(self, cet_tar, domain, coef):
        self.cet_tar = cet_tar
        self.center = cet_tar[0]
        self.target = cet_tar[1]
        self.domain = domain
        self.coef = coef

    def to_str(self):
        cfx = self.coef[0,:,:]
        cfy = self.coef[1,:,:]
        cfz = self.coef[2,:,:]
        cf_xyz = np.concatenate((cfx,cfy,cfz))
        cf_str = num2txt(cf_xyz)
        n_cols = cf_xyz.shape[1]
        ini_dom = self.domain[0,0]
        dt_rec = self.domain[0,1] - self.domain[0,0]
        dt_dom = self.domain[-1,-1] - self.domain[0,0]
        n_recs = cfx.shape[0]

        first_row = [self.center, self.target, n_cols, n_recs, dt_rec, ini_dom, dt_dom]
        first_row = ['segment'] + [str(int(i)) for i in first_row]
        first_row = ','.join(first_row)+'\n'
        return first_row + cf_str

    def get_pos(self, t):
        """
        Get position of the object at time t
        
        Parameters
        ----------
            t (datetime)    : time for which the position is requested
        
        Returns
        ----------
            pos (np.array): position of the object at t
        """
        jd = datetime_to_jd(t)
        t = jd_to_sec(jd)
        mask = np.logical_and(t>=self.domain[:,0], t<self.domain[:,1])
        rec = np.where(mask)[0][0] # record index
        cfx = self.coef[0,rec,:]
        cfy = self.coef[1,rec,:]
        cfz = self.coef[2,rec,:]
        fx = np.polynomial.chebyshev.Chebyshev(coef=cfx, domain=self.domain[rec])
        fy = np.polynomial.chebyshev.Chebyshev(coef=cfy, domain=self.domain[rec])
        fz = np.polynomial.chebyshev.Chebyshev(coef=cfz, domain=self.domain[rec])
        pos = np.vstack((fx(t),fy(t),fz(t))).T[0]
        return pos


def load_pickle(fname):
    """
    Load an ephemeris pickle file (created by numeph python package)
    
    Parameters
    ----------
        fname (str) : path and name of the pickle file
    
    Returns
    ----------
        Dictionary of Segments
    """
    f = open(fname, 'rb')
    data = pickle.load(f)
    f.close()
    dc = {}
    for k,v in data.items():
        dc[k] = Segment(k, v[0], v[1])
    return dc



"""
Module cosmology
===============
This module has a parent classes (CosModel) to represent a Flat Lambda-CDM
model of the universe.
There are also child classes (suck as Plank18) who provide initial data for
the model.
"""

import numpy as np
import scipy.integrate as integrate

class CosModel:
    """
    Cosmological model by density parameters (Omegas)

    Parameters
    ----------
        H0 (float)  : Current Hubble constant
        Oph0 (float): Current density parameter of photon
        One0 (float): Current density parameter of neutrino
        Obm0 (float): Current density parameter of baryonic matter
        Odm0 (float): Current density parameter of dark matter
        Ode0 (float): Current density parameter of dark energy

    Attributes
    ----------
        sf (np.poly1d): scale factor
        
    Methods
    -------
        scale_factor: Scale factor
        current_age: Current age of the universe
        proper_distance: Proper distance between two moments, observed & emitted
        z: The redshift
    """
    def __init__(self, H0, Oph0, One0, Obm0, Odm0, Ode0):
        
        self.H0 = H0
        self.Oph0 = Oph0
        self.One0 = One0
        self.Obm0 = Obm0
        self.Odm0 = Odm0
        self.Ode0 = Ode0
        self.sf = None

    def initial_calculations(self):
        # Convert unit of H0 from ('km / (Mpc s)') to ('1 / Gyr')
        self.H0 = self.H0 * 0.00102271
        # Density parameters
        self.Or0 = self.Oph0 + self.One0
        self.Om0 = self.Obm0 + self.Odm0
        self.Oall = self.Or0 + self.Om0 + self.Ode0

    def scale_factor(self, a_min=1e-10, a_max=5, n_points=3000, fit_deg=9):
        """
        Scale factor of the universe in terms of time.
        
        Arguments
        ----------
            a_min (float): maximum value for scale factor
            a_max (float): minimum value for scale factor
            n_points (int): number of points used to fit
            fit_deg (int): degree of the polynomiat fit
        
        Returns
        -------
            Scale factor as a np.poly1d object
        """
        if self.sf is None:
            self.initial_calculations()
            a = np.linspace(a_min, a_max, n_points)
            tmp = lambda a: 1 / np.sqrt(
                self.Or0/a**2 + self.Om0/a + self.Ode0*a**2 + 1-self.Oall
                )
            t = np.array([integrate.quad(tmp, 0, i)[0]/self.H0 for i in a])
            coefs = np.polyfit(t, a, fit_deg)
            self.sf = np.poly1d(coefs)
        return self.sf

    def current_age(self):
        """
        Current age of the universe

        Returns
        -------
            Current age as Billion years from the Big bang
        """
        if self.sf is None:
            self.sf = self.scale_factor()
        r = np.roots(self.sf-1)
        return r[np.isreal(r)].real[0]

    def proper_distance(self, to, te):
        """
        Proper distance between two moments, observed (to) and emitted (te)
        
        Arguments
        ----------
            to (float): emitted moment (in Billion years from Big bang)
            te (float): observed moment (in Billion years from Big bang)
        
        Returns
        -------
            Proper distance in Billion light-years
        """
        if self.sf is None:
            self.sf = self.scale_factor()
        dp = integrate.quad(lambda t: 1/self.sf(t), te, to)[0]
        return dp

    def z(self, t):
        """
        Redshift (z value) at a given moment
        
        Arguments
        ----------
            t (float): moment (in Billion years from Big bang)
        
        Returns
        -------
            z value as float
        """
        if self.sf is None:
            self.sf = self.scale_factor()
        return (1/self.sf(t))-1


class Planck18(CosModel):
    """
    Predefined CosModel class with density parameters provided by Planck 2018
    """
    def __init__(self):
        self.H0 = 67.66
        self.Oph0 = 5.402015137139351e-05
        self.One0 = 0.0014396743040845238
        self.Obm0 = 0.04897
        self.Odm0 = 0.26069
        self.Ode0 = 0.6888463055445441
        self.sf = np.poly1d([ 3.14416422e-17, -8.77434127e-15, 1.09165168e-12,
            -7.98435601e-11, 3.81160114e-09, -1.24683679e-07, 2.85512434e-06,
            -4.59502352e-05, 5.14673404e-04, -3.92192424e-03, 1.96940133e-02,
            -6.30581943e-02, 1.85807439e-01,  1.54295181e-02])



"""
Module catalogues
=================
This module retrieves data from different catalogues via VizieR TAP Service.
You can find the current available catalogues with the dict:

>>> from hypatie import available_catalogues
>>> print(available_catalogues)
{'lqac5': 'Large Quasar Astrometric Catalogue (5th release)',
 'ucac4': 'U.S. Naval Observatory CCD Astrograph Catalog (4th release)',
 'hipparcos': 'HIgh Precision PARallax COllecting Satellite (Hipparcos)',
 'sdss12': 'SDSS Photometric Catalogue, (Release 12)',
 'glimpse': 'GLIMPSE Source Catalog (I + II + 3D) (IPAC 2008)',
 'gaia2': 'Gaia data release 2',
 'gaiae3': 'Gaia early data release 3'}
"""
from urllib.request import urlopen
import pandas as pd
import json

BASE = """http://tapvizier.u-strasbg.fr/TAPVizieR/tap/sync?\
request=doQuery&lang=adql&format=json&query=""".replace(' ','')

# Catalogue names
lqac5_nm = 'Large Quasar Astrometric Catalogue (5th release)'
ucac4_nm = 'U.S. Naval Observatory CCD Astrograph Catalog (4th release)'
hipparcos_nm = 'HIgh Precision PARallax COllecting Satellite (Hipparcos)'
sdss12_nm = 'SDSS Photometric Catalogue, (Release 12)'
glimpse_nm = 'GLIMPSE Source Catalog (I + II + 3D) (IPAC 2008)' #GLIMPSE sources in Galactic Center (104240613 rows)
gaia2_nm = 'Gaia data release 2'
gaiae3_nm = 'Gaia early data release 3'
dfgrs2_nm = '2dF Galaxy Redshift Survey'
glade2_nm = 'GLADE v2.3 catalog (Galaxy List for the Advanced Detector Era)'

# Table names
lqac5_tb = 'J/A%2bA/624/A145/lqac5'
ucac4_tb = 'I/322A/out'
hipparcos_tb = 'I/239/hip_main'
sdss12_tb = 'V/147/sdss12'
glimpse_tb = 'II/293/glimpse'
gaia2_tb = 'I/345/gaia2'
gaiae3_tb = 'I/350/gaiaedr3'
dfgrs2_tb = 'VII/250/2dfgrs'
glade2_tb = 'VII/281/glade2'

# Field names
lqac5_fn = ['LQAC','RAJ2000','DEJ2000','Plx','pmRA*','pmDE','Ref','GAIA',
            'z','Gmag2','BPmag','RPmag','umag','Bmag','Vmag','gmag',
            'rmag','imag','zmag','Jmag','Kmag','W1mag','W2mag','W3mag',
            'W4mag','S1.4','S2.3','S5.0','S8.4','S23','BMAG','IMAG','uMAG',
            'gMAG','rMAG','iMAG','zMAG']
ucac4_fn = ['UCAC4','RAJ2000','DEJ2000','pmRA','pmDE','2Mkey','f.mag',
            'a.mag','Jmag','Hmag','Kmag','Bmag','Vmag','gmag','rmag',
            'imag','H']
hipparcos_fn = ['HIP','_RA.icrs','_DE.icrs','Vmag','Plx','e_Plx',
                'pmRA','pmDE','B-V','Period','HvarType','SpType']
sdss12_fn = ['RA_ICRS','DE_ICRS','class','objID','SDSS12','ObsDate','Q',
             'zsp','e_zsp','zph','e_zph','<zph>','spType','spCl']
glimpse_fn = ['S','GLIMPSE','2MASS','Glon','Glat','RAJ2000','DEJ2000',
              'Jmag','Hmag','Kmag','3.6mag','4.5mag','5.8mag','8.0mag']
gaia2_fn = ['designation','source_id','random_index','ra_epoch2000',
            'dec_epoch2000','l','b','parallax','lum_val','radius_val',
            'pmra','pmdec','radial_velocity','teff_val','phot_rp_mean_mag',
            'phot_g_mean_mag','phot_bp_mean_mag','phot_rp_mean_flux',
            'phot_g_mean_flux','phot_bp_mean_flux','a_g_val',
            'phot_bp_rp_excess_factor']
gaiae3_fn = ['EDR3Name','RAJ2000','DEJ2000','Epoch','Source','RandomI',
             'Plx','e_Plx','pmRA','pmDE','NAL','NAC','MatchObs','FG','e_FG',
             'FGCorr','Gmag','e_Gmag','GmagCorr','FBP','e_FBP','BPmag',
             'e_BPmag','FRP','e_FRP','RPmag','e_RPmag','RVDR2','e_RVDR2',
             'Tefftemp','GLON','GLAT']
dfgrs2_fn = ['SeqNum','Name','RAJ2000','DEJ2000','Bjmag','Bjsel','z','q_z',
             'n_z','z.em','o_z.em','SNR']
glade2_fn = ['PGC','GWGC','HyperLEDA','2MASS','SDSS-DR12','Flag1','RAJ2000',
             'DEJ2000','Dist','e_Dist','z','Bmag','e_Bmag','BMAG','Jmag',
             'e_Jmag','Hmag','e_Hmag','Kmag','e_Kmag','Flag2','Flag3']

dc = {'lqac5'     : [lqac5_tb, lqac5_fn],
      'ucac4'     : [ucac4_tb, ucac4_fn],
      'hipparcos' : [hipparcos_tb, hipparcos_fn],
      'sdss12'    : [sdss12_tb, sdss12_fn],
      'glimpse'   : [glimpse_tb, glimpse_fn],
      'gaia2'     : [gaia2_tb, gaia2_fn],
      'gaiae3'    : [gaiae3_tb, gaiae3_fn],
      'dfgrs2'    : [dfgrs2_tb, dfgrs2_fn],
      'glade2'    : [glade2_tb, glade2_fn]}

long_names = [lqac5_nm, ucac4_nm, hipparcos_nm, sdss12_nm,
              glimpse_nm, gaia2_nm, gaiae3_nm, dfgrs2_nm, glade2_nm]

# dictionary containing names of available catalogues
available_catalogues = {k:v for (k,v) in [*zip(dc.keys(), long_names)]}

class Catalogue:
    """
    Catalogue class

    Parameters
    ----------
        name (str)     : name of the catalogue
        columns (list) : columns of data to be requested
        where (str)    : condition (WHERE clause in SQL)
        order_by (str) : column to order (ORDER BY clause in SQL)
        n_max (int)    : maximum number of rows to return; dfault 100.
        random_sampling (bool): return a rondom sample of data.
            Note that huge datasets can create Gateway timeout error
            using this parameter; try to reduce the size of data using
            'where' in the case of Gateway timeout error.

    Attributes
    ----------
        tbl_name        : name of the table
        columns         : columns of the table
        random_sampling : whether or not 'random_sampling' is True
        sql             : ADQL script to produce data
        url             : URL of VizieR to produce data

    Methods
    -------
        available_columns : return all available possible columns in the
                            table of the catalogue
        download : download data from the catalogue
    """
    def __init__(self,
                 name,
                 columns=None,
                 where=None,
                 order_by=None,
                 n_max=100,
                 random_sampling=False):
        
        self.tbl_name = dc[name][0]
        self.random_sampling = random_sampling
        
        if columns is None:
            self.columns = dc[name][1]
        elif isinstance(columns, list):
            self.columns = columns
        elif columns=='all':
            self.columns = list(self.available_columns().index)

        if where is not None:
            where_words = where.split(' ')
            where_cols = [i for i in where_words if i in \
                          list(self.available_columns().index)]
            if len(where_cols)==0:
                raise Exception("column(s) defined is 'where' is/are not valid!")
            _where = []
            for word in where_words:
                if word in where_cols:
                    _where.append('"'+word+'"')
                else:
                    _where.append(word)
            where = ' WHERE ' + ' '.join(_where)
        else:
            where = ''

        if order_by is None:
            order_by = ''
        else:
            if ' desc' in order_by.lower():
                order_by = ' ORDER BY "'+order_by[:-5]+'"' + " DESC"
            else:
                order_by = ' ORDER BY "'+order_by+'"'
        
        if self.random_sampling and isinstance(n_max, int):
            rnd = 'RAND(1) as rnd_sample,'
            order_by = ' ORDER BY rnd_sample'
        else:
            rnd = ''
        if isinstance(n_max, int):
            n_max = f" TOP {n_max} "
        elif n_max=='all':
            n_max = ''
        else:
            raise Exception("'n_max' should be integer or 'all'")

        sql_columns = [f'"{i}"' for i in self.columns]
        self.sql = 'SELECT' + n_max + rnd + ','.join(sql_columns) + \
                   ' FROM "' + self.tbl_name+'"' + where + order_by
        self.url = (BASE+self.sql).replace(' ', '%20')

    def available_columns(self):
        """
        Get all available columns in the main table of the catalogue

        Returns
        -------
            pandas DataFrame of available columns and descriptions

        Ex:
        >>> cat = Catalogue('gaia2')
        >>> print(cat.available_columns())
        """
        
        sql = 'SELECT column_name, description, unit FROM tap_schema.columns'
        sql = sql + f" WHERE table_name='{self.tbl_name}'"
        url = (BASE+sql).replace(' ', '%20')
        r = json.loads(urlopen(url).read().decode('utf-8'))
        cols = [i[0] for i in r['data']]
        desc = [i[1] for i in r['data']]
        df = pd.DataFrame(r['data'],
                columns=['column','description','unit'])
        return df.set_index('column')

    def download(self):
        """
        Download data from the main table of the catalogue

        Returns
        -------
            data : pandas DataFrame of the main data table
            meta : pandas DataFrame of the meta-data table

        Ex:
        >>> cat = Catalogue('gaia2')
        >>> data, meta = cat.download()
        """
        r = json.loads(urlopen(self.url).read().decode('utf-8'))
        cols = [i['name'] for i in r['metadata']]
        data = pd.DataFrame(r['data'], columns=cols)
        meta = pd.DataFrame(r['metadata'])
        if self.random_sampling:
            data = data.drop('rnd_sample', axis=1)
            meta = meta.drop([0])
        meta = meta[['name', 'description', 'unit']].set_index('name')
        return data, meta


